import sys
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, List, Optional, Union

import better_exceptions  # type:ignore

from alarmer.provider import Provider
from alarmer.throttling import Throttling


class Alarmer:
    _old_except_hook: Callable
    _providers: List[Union[Provider, Callable]]
    _pool: ThreadPoolExecutor
    _global_throttling: Optional[Throttling]
    environment: str

    @classmethod
    def except_hook(cls, exc, value, tb):
        context = None
        if tb is not None:
            prev = tb
            curr = tb.tb_next
            while curr is not None:
                prev = curr
                curr = curr.tb_next
            context = prev.tb_frame.f_locals
        cls._old_except_hook(exc, value, tb)
        message = "".join(better_exceptions.format_exception(exc, value, tb))
        cls.send(message, exc, context)

    @classmethod
    def send(
        cls, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None
    ):
        for p in cls._providers:
            if isinstance(p, Provider):
                t = p.throttling or cls._global_throttling
                if (t and t(p, message, exc, context)) or not t:
                    cls._pool.submit(p.send, message, exc, context)
            else:
                if cls._global_throttling and cls._global_throttling(p, message, exc, context):
                    cls._pool.submit(p, message, exc, context)

    @classmethod
    def init(
        cls,
        providers: List[Union[Provider, Callable]],
        thread_pool_size: Optional[int] = None,
        global_throttling: Optional[Throttling] = Throttling(),
        environment: str = "production",
    ):
        better_exceptions.MAX_LENGTH = None
        cls._global_throttling = global_throttling
        cls._old_except_hook = sys.excepthook
        cls.environment = environment
        cls._providers = providers
        sys.excepthook = cls.except_hook
        cls._pool = ThreadPoolExecutor(max_workers=thread_pool_size)



from typing import Optional

import requests

from alarmer import Provider, Throttling


class WeComProvider(Provider):
    """
    https://work.weixin.qq.com/api/doc/90000/90136/91770
    """

    def __init__(
        self,
        webhook_url: str,
        throttling: Optional[Throttling] = None,
    ):
        super().__init__(throttling)
        self.webhook_url = webhook_url

    def send(
        self, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None
    ):
        data = {"msgtype": "text", "content": {"text": message}}
        return requests.post(
            self.webhook_url,
            json=data,
        )



import abc
from typing import Optional

from alarmer.throttling import Throttling


class Provider(abc.ABC):
    def __init__(self, throttling: Optional[Throttling] = None):
        self.throttling = throttling

    def send(
        self, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None
    ):
        raise NotImplementedError()



import json
from typing import Optional

import requests

from alarmer import Alarmer, Provider, Throttling
from alarmer.encoder import ComplexEncoder


class FeiShuProvider(Provider):
    """
    https://www.feishu.cn/hc/zh-CN/articles/360024984973
    """

    def __init__(
        self,
        webhook_url: str,
        throttling: Optional[Throttling] = None,
    ):
        super().__init__(throttling)
        self.webhook_url = webhook_url

    def send(
        self, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None
    ):
        data: dict = {
            "msg_type": "interactive",
            "card": {
                "header": {
                    "title": {
                        "content": f"[{Alarmer.environment}] Exception Alarm: {exc}"
                        if exc
                        else f"[{Alarmer.environment}] Alarm",
                        "tag": "plain_text",
                    },
                    "template": "red",
                },
                "elements": [
                    {
                        "tag": "div",
                        "text": {"content": message, "tag": "plain_text"},
                    },
                ],
            },
        }
        if context:
            context_content = json.dumps(context, indent=4, cls=ComplexEncoder)
            data["card"]["elements"].extend(
                [
                    {"tag": "hr"},
                    {
                        "tag": "div",
                        "fields": [
                            {
                                "is_short": True,
                                "text": {"tag": "lark_md", "content": "**Context Variables:**"},
                            },
                            {
                                "is_short": False,
                                "text": {"tag": "plain_text", "content": context_content},
                            },
                        ],
                    },
                ]
            )
        return requests.post(
            self.webhook_url,
            json=data,
        )



import logging

from alarmer import Alarmer


class LoggingHandler(logging.Handler):
    def __init__(self, level: int = logging.ERROR):
        super().__init__()
        self.level = level

    def emit(self, record: logging.LogRecord) -> None:
        if record.levelno >= self.level:
            exc_info = record.exc_info
            exc = None
            if exc_info:
                exc = exc_info[1]
            Alarmer.send(record.getMessage(), exc)



from typing import List, Optional

import apprise

from alarmer import Alarmer, Throttling
from alarmer.provider import Provider


class AppriseProvider(Provider):
    def __init__(
        self,
        services: List[str],
        throttling: Optional[Throttling] = None,
        apprise_options: Optional[dict] = None,
        notify_options: Optional[dict] = None,
    ):
        super().__init__(throttling)
        self.app = apprise.Apprise(services, **(apprise_options or {}))
        self.notify_options = notify_options or {}

    def send(
        self, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None
    ):
        title = (
            f"[{Alarmer.environment}] Exception Alarm: {exc}"
            if exc
            else f"[{Alarmer.environment}] Alarm"
        )

        return self.app.notify(title=title, body=message, **self.notify_options)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['alarmer', 'alarmer.provider']

package_data = \
{'': ['*']}

install_requires = \
['apprise', 'better-exceptions']

setup_kwargs = {
    'name': 'alarmer',
    'version': '0.1.4',
    'description': 'A tool focus on error reporting for your application',
    'long_description': '# alarmer\n\n[![image](https://img.shields.io/pypi/v/alarmer.svg?style=flat)](https://pypi.python.org/pypi/alarmer)\n[![image](https://img.shields.io/github/license/long2ice/alarmer)](https://github.com/long2ice/alarmer)\n[![pypi](https://github.com/long2ice/alarmer/actions/workflows/pypi.yml/badge.svg)](https://github.com/long2ice/alarmer/actions/workflows/pypi.yml)\n[![ci](https://github.com/long2ice/alarmer/actions/workflows/ci.yml/badge.svg)](https://github.com/long2ice/alarmer/actions/workflows/ci.yml)\n\n`Alarmer` is a tool focus on error reporting for your application, like [sentry](https://sentry.io) but lightweight.\n\n## Installation\n\n```shell\npip install alarmer\n```\n\n## Usage\n\nIt\'s simple to integrate `alarmer` in your application, just call `Alarmer.init` on startup of your application.\n\n```py\nimport os\n\nfrom alarmer import Alarmer\nfrom alarmer.provider.feishu import FeiShuProvider\n\n\ndef main():\n    Alarmer.init(providers=[FeiShuProvider(webhook_url=os.getenv("FEI_SHU_WEBHOOK_URL"))])\n    raise Exception("test")\n\n\nif __name__ == "__main__":\n    main()\n```\n\n### Intercept Error Logging\n\nIf you want to intercept the logging, you can use `LoggingHandler`.\n\n```py\nimport logging\nimport os\n\nfrom alarmer import Alarmer\nfrom alarmer.log import LoggingHandler\nfrom alarmer.provider.feishu import FeiShuProvider\n\n\ndef main():\n    Alarmer.init(providers=[FeiShuProvider(webhook_url=os.getenv("FEI_SHU_WEBHOOK_URL"))])\n    logging.basicConfig(\n        level=logging.INFO,\n    )\n    logger = logging.getLogger()\n    logger.addHandler(LoggingHandler(level=logging.ERROR))  # only error and above should be send\n    logging.error("test logging")\n\n\nif __name__ == "__main__":\n    main()\n\n```\n\nNow when you run the script, you will receive the errors in your provider.\n\n## Provider\n\nYou can set number of providers for error reporting. All kinds of providers can be found\nin [providers](./alarmer/provider).\n\nThanks to [Apprise](https://github.com/caronc/apprise), you can use lots of providers out of box.\n\n- [Apprise](https://github.com/caronc/apprise)\n- [FeiShu](https://www.feishu.cn/hc/zh-CN/articles/360024984973)\n- [WeCom](https://work.weixin.qq.com/api/doc/90000/90136/91770)\n\n### Custom Provider\n\nYou can write your own custom provider by inheriting the `Provider` class.\n\n```py\nfrom typing import Optional\nfrom alarmer.provider import Provider\n\n\nclass CustomProvider(Provider):\n\n    def send(self, message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None):\n        # Send to your custom provider here\n        pass\n```\n\nIn addition to this, you can just write a callable function which takes `message` and `exc` arguments.\n\n```py\nfrom typing import Optional\n\n\ndef custom_provider(message: str, exc: Optional[BaseException] = None, context: Optional[dict] = None):\n    # Send to your custom provider here\n    pass\n```\n\nThen add it to `Alarmer.init`.\n\n```py\nfrom alarmer import Alarmer\n\nAlarmer.init(providers=[CustomProvider(), custom_provider])\n```\n\n## Throttling\n\n`Throttling` is used to throttling error messages if there are too many errors.\n\n```py\nfrom alarmer import Alarmer\nfrom alarmer.throttling import Throttling\n\nAlarmer.init(global_throttling=Throttling(), providers=[...])\n```\n\n### Custom Throttling\n\nYou can write your own throttling by inheriting the `Throttling` class.\n\n```py\nimport typing\n\nfrom alarmer.throttling import Throttling\n\nif typing.TYPE_CHECKING:\n    from alarmer.provider import Provider\n\n\nclass MyThrottling(Throttling):\n    def __call__(self, provider: "typing.Union[Provider,typing.Callable]", message: str,\n                 exc: typing.Optional[BaseException] = None, context: typing.Optional[dict] = None) -> bool:\n        # check whether the error message should be send\n        return True\n```\n\n## Manual Send\n\nIf you want to manually send messages to the providers, just call `Alarmer.send`.\n\n```py\nfrom alarmer import Alarmer\n\nAlarmer.send("message")\n```\n\n## License\n\nThis project is licensed under the\n[Apache-2.0](https://github.com/long2ice/alarmer/blob/master/LICENSE) License.\n',
    'author': 'long2ice',
    'author_email': 'long2ice@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/long2ice/alarmer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Any


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj: Any):
        if not isinstance(obj, (int, float, bool)):
            return str(obj)
        return super().default(obj)



import threading
import time
import typing

if typing.TYPE_CHECKING:
    from alarmer import Provider


class Throttling:
    def __init__(self):
        self.last_time = time.time()
        self.lock = threading.Lock()

    def __call__(
        self,
        provider: "typing.Union[Provider,typing.Callable]",
        message: str,
        exc: typing.Optional[BaseException] = None,
        context: typing.Optional[dict] = None,
    ) -> bool:
        with self.lock:
            if time.time() - self.last_time < 5:
                return False
            self.last_time = time.time()
            return True
